import xarray as xr
import numpy as np
from math import nan

from . import Utils


# Internal functions:


def _mono_window(bt, band_lambda, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bt: (np.ndarray|xr.Dataset): band at ~10-12 um in BT calibration
        band_lambda: (float) wavelength of emitted radiance in used band
            (the peak response or average of the limiting wavelength)
            Unit is um
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    bt_c = bt - 273.15  # to Celsius

    ndvi_s = 0.2
    ndvi_v = 0.5
    p_v = ((ndvi - ndvi_s) / (ndvi_v - ndvi_s))**2

    c = 0.005  # represents surface roughness

    e_s = 0.966  # soil emissivity
    e_v = 0.973  # vegetation emissivity
    e_w = 0.991  # water emmisivity

    e_l = e_v*p_v + e_s*(1-p_v) + c
    e_l = xr.where(ndvi < ndvi_s, e_s, e_l)
    e_l = xr.where(ndvi < 0, e_w, e_l)  # ndvi < 0 indicates water
    e_l = xr.where(ndvi > ndvi_v, e_v, e_l)

    p = 1.438e-2

    lst = bt_c / (1 + (band_lambda*bt_c/p) * np.log(e_l) * 1e-12)
    if cmask is not None:
        lst = xr.where(cmask == 1, lst, nan)
    return lst


# Private wrappers

def _mono_window_i05(bi05, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bi05: (np.ndarray|xr.Dataset): I05 in BT calibration
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    return _mono_window(bi05, (10.5+12.4)/2, ndvi, cmask=cmask)


def _mono_window_m15(bm15, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bm15: (np.ndarray|xr.Dataset): M15 in BT calibration
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    return _mono_window(bm15, (10.263+11.263)/2, ndvi, cmask=cmask)


def _mono_window_m16(bm16, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bm16: (np.ndarray|xr.Dataset): M16 in BT calibration
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    return _mono_window(bm16, (11.538+12.488)/2, ndvi, cmask=cmask)


# Public wrappers:


def mono_window_i05(bi05, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bi05: (np.ndarray|xr.Dataset): I05 in BT calibration
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    Utils._check_data(bi05)
    Utils._check_data(ndvi)
    if cmask is not None:
        Utils._check_data(cmask)
    return _mono_window_i05(
        bi05, ndvi,
        cmask=cmask
    )


def mono_window_m15(bm15, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bm15: (np.ndarray|xr.Dataset): M15 in BT calibration
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    Utils._check_data(bm15)
    Utils._check_data(ndvi)
    if cmask is not None:
        Utils._check_data(cmask)
    return _mono_window_m15(
        bm15, ndvi,
        cmask=cmask
    )


def mono_window_m16(bm16, ndvi, cmask=None):
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bm15: (np.ndarray|xr.Dataset): M16 in BT calibration
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    Utils._check_data(bm16)
    Utils._check_data(ndvi)
    if cmask is not None:
        Utils._check_data(cmask)
    return _mono_window_m16(
        bm16, ndvi,
        cmask=cmask
    )


# Public xr.Dataset wrappers:


def mono_window_i05_ds(ds, ndvi, cmask=None):
    """
        Wrapper for LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        ds (xr.Dataset): dataset with I05 band data
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return mono_window_i05(
        ds['I05'], ndvi,
        cmask=cmask
    )


def mono_window_m15_ds(ds, ndvi, cmask=None):
    """
        Wrapper for LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        ds (xr.Dataset): dataset with M15 band data
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return mono_window_m15(
        ds['M15'], ndvi,
        cmask=cmask
    )


def mono_window_m16_ds(ds, ndvi, cmask=None):
    """
        Wrapper for LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        ds (xr.Dataset): dataset with M16 band data
        ndvi: (np.ndarray|xr.Dataset): NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return mono_window_m16(
        ds['M16'], ndvi,
        cmask=cmask
    )



import xarray as xr
from math import nan

from . import Utils


# Internal functions:


def _rsnpp_day_img(ri1, ri2, ri3, bi4, bi5):
    """
        Day reflectance/thermal I-bands cloud test
        Based on the M.Piper, T.Bahr (2015).
        A RAPID CLOUD MASK ALGORITHM FOR SUOMI NPP VIIRS IMAGERY EDRS
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
    Returns:
        (np.ndarray|xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """

    t1 = xr.where(ri1 > 8, True, False)

    ndsi = (ri1 - ri3) / (ri1 + ri3)
    t2 = xr.where(ndsi < 0.7, True, False)
    t2 = xr.where(ri2 > 11, t2, False)

    t3 = xr.where(bi5 < 300, True, False)

    ri3_max = Utils.max_values(ri3)
    t4 = xr.where((ri3_max - ri3) * bi5 / 100 < 410, True, False)

    t5 = xr.where(ri2 / ri1 < 2, True, False)

    t6 = xr.where(ri2 / ri3 > 1, True, False)

    cm = xr.where(t1 & t2, 0, 1)
    cm = xr.where(cm & t3, 0, 1)
    cm = xr.where(cm & t4, 0, 1)
    cm = xr.where(cm & t5, 0, 1)
    cm = xr.where(cm & t6, 0, 1)

    nanmask = Utils.nan_mask(ri1)
    cm = xr.where(nanmask, nan, cm)
    return cm


def _fire_day_img(ri1, ri2, bi5):
    """
        Day reflectance & termal I-bands cloud test
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
    Returns:
        (np.ndarray|xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    cm = xr.where(bi5 < 265, 0, 1)

    sum_ri = ri1 + ri2

    t2 = xr.where(sum_ri > 90, 0, 1)
    t2 = xr.where(bi5 < 295, t2, 1)

    t3 = xr.where(sum_ri > 70, 0, 1)
    t3 = xr.where(bi5 < 285, t2, 1)

    cm = xr.where(t2 == 0, 0, cm)
    cm = xr.where(t3 == 0, 0, cm)

    cm = xr.where(ri1 is nan, nan, cm)
    return cm


def _fire_night_img(bi4, bi5, nmask=None):
    """
        Night termal I-bands cloud test
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
        nmask (np.ndarray|xr.Dataset, optional): Day/night mask (1 is night)
    Returns:
        (np.ndarray|xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    cm = xr.where(bi4 < 265, 0, 1)
    cm = xr.where(bi5 < 295, cm, 0)
    cm = xr.where(Utils.not_nan_mask(bi4), cm, nan)

    if nmask is not None:
        cm = xr.where(nmask == 1, cm, nan)
    else:
        cm = xr.where(bi4 is nan, nan, cm)
    return cm


# Public wrappers:


def rsnpp_day_img(ri1, ri2, ri3, bi4, bi5):
    """
        Day reflectance/thermal I-bands cloud test
        Based on the M.Piper, T.Bahr (2015).
        A RAPID CLOUD MASK ALGORITHM FOR SUOMI NPP VIIRS IMAGERY EDRS
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
    Returns:
        (np.ndarray|xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    Utils._check_data(ri1)
    Utils._check_data(ri2)
    Utils._check_data(ri3)
    Utils._check_data(bi4)
    Utils._check_data(bi5)

    return _rsnpp_day_img(
        ri1, ri2, ri3,
        bi4, bi5
    )


def fire_day_img(ri1, ri2, bi5):
    """
        Day reflectance & termal I-bands cloud test
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
    Returns:
        (np.ndarray|xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    Utils._check_data(ri1)
    Utils._check_data(ri2)
    Utils._check_data(bi5)
    return _fire_day_img(ri1, ri2, bi5)


def fire_night_img(bi4, bi5, nmask=None):
    """
        Night termal I-bands cloud test
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
        nmask (np.ndarray|xr.Dataset, optional): Day/night mask (1 is night)
    Returns:
        (np.ndarray|xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    Utils._check_data(bi4)
    Utils._check_data(bi5)
    return _fire_night_img(bi4, bi5, nmask=nmask)


# Public xr.Dataset wrappers:


def rsnpp_day_img_ds(ds):
    """
        Wrapper for day reflectance/thermal I-bands cloud test
            for xr.Dataset objects
        Based on the M.Piper, T.Bahr (2015).
        A RAPID CLOUD MASK ALGORITHM FOR SUOMI NPP VIIRS IMAGERY EDRS
    Args:
        ds (xr.Dataset): dataset with I01-I05 bands data
    Returns:
        (xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return rsnpp_day_img(
        ds['I01'],
        ds['I02'],
        ds['I03'],
        ds['I04'],
        ds['I05']
    )


def fire_day_img_ds(ds):
    """
        Wrapper for Day reflectance & termal I-bands cloud test for xr.Dataset
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ds (xr.Dataset): dataset with I01, I02, I05 bands data
    Returns:
        (xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return fire_day_img(
        ds['I01'],
        ds['I02'],
        ds['I05']
    )


def fire_night_img_ds(ds, nmask=None):
    """
        Wrapper for Night termal I-bands cloud test for xr.Dataset
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ds (xr.Dataset): dataset with I04, I05 bands data
        nmask (np.ndarray|xr.Dataset, optional): Day/night mask (1 is night)
    Returns:
        (xr.Dataset): integer cloud mask,
            0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return fire_night_img(
        ds['I04'],
        ds['I05'],
        nmask=nmask
    )



import xarray as xr
import math

from . import Utils

# Public functions:


def naive(refband, btband):
    """
        Get night mask from any reflectance and brightness bands
        Day/Night state here meets the condition SZA < 90 deg
        Assumed that data was loaded in the reflectance or
            brightness-temperature, not radiance calibration
    Args:
        refband (np.ndarray|xr.DataArray):
            any reflectance data
        btband (np.ndarray|xr.DataArray):
            any brightness-temperature data
    Returns:
        (np.ndarray|xr.DataArray): integer mask,
            1 means night state, 0 means day state,
            Can contain NaN values in case of missing data in BT band
    """
    bmask = Utils.not_nan_mask(btband)
    rmask = Utils.nan_mask(refband)
    return xr.where(bmask, rmask, math.nan).astype(float)


# Internal wrappers:


def _naive_ds(ds, refbands, btbands):
    """
        Internal wrapper for getting night mask from xr.Dataset
        Day/Night state here meets the condition SZA < 90 deg
        Assumed that data was loaded in the reflectance or
            brightness-temperature, not radiance calibration
    Args:
        ds (xr.Dataset): original data
        refbands (list of strings): names of reflectance datasets
        btbands (list of strings): names of BT datasets
    Returns:
        (xr.DataArray): integer mask,
            1 means night state, 0 means day state,
            Can contain NaN values in case of missing data in BT band
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    refband = None
    btband = None
    for b in refbands:
        if b in ds.data_vars:
            refband = b
            break
    for b in btbands:
        if b in ds.data_vars:
            btband = b
            break
    if refband is None or btband is None:
        raise ValueError(
            "Input data does not contain both ref and bt data"
        )
    rd = Utils._get_da(ds, refband)
    bd = Utils._get_da(ds, btband)
    return naive(rd, bd)


# Public xr.Dataset wrappers:


def naive_ds_img(ds):
    """
        Wrapper for getting night mask from xr.Dataset directly
            using imagery resolution (I-bands)
        Day/Night state here meets the condition SZA < 90 deg
        Assumed that data was loaded in the reflectance or
            brightness-temperature, not radiance calibration
    Args:
        ds (xr.Dataset): contains both reflectance and
            brightness-temperature data
    Returns:
        (xr.DataArray): integer mask,
            1 means night state, 0 means day state,
            Can contain NaN values in case of missing data in BT band
    """
    rbands = ['I{:02d}'.format(i) for i in range(1, 3 + 1)]
    bbands = ['I{:02d}'.format(i) for i in range(4, 5 + 1)]
    return _naive_ds(ds, rbands, bbands)


def naive_ds_mod(ds):
    """
        Wrapper for getting night mask from xr.Dataset directly
            using moderate resolution (M-bands)
        Day/Night state here meets the condition SZA < 90 deg
        Assumed that data was loaded in the reflectance or
            brightness-temperature, not radiance calibration
    Args:
        ds (xr.Dataset): contains both reflectance and
            brightness-temperature data
    Returns:
        (xr.DataArray): integer mask,
            1 means night state, 0 means day state,
            Can contain NaN values in case of missing data in BT band
    """
    rbands = ['M{:02d}'.format(i) for i in range(1, 11 + 1)]
    bbands = ['M{:02d}'.format(i) for i in range(12, 16 + 1)]
    return _naive_ds(ds, rbands, bbands)



from netCDF4 import Dataset  # require netcdf4 being installed, not NetCDF4

from . import ReadingHelpers as rh


def read_npp_viaes_l1(path):
    """
        Read VIIRS I-band imagery product (VIAES_L1)
    Args:
        path (string): path to the desired file
    Returns:
        dict: datasets with reflectance, radiance,
            brightness temperature data
            as masked np-arrays
    """
    data = {}
    with Dataset(path, 'r') as ifile:
        # get reflectance data
        for i in range(1, 4):
            data[f'refi{i}'] = rh.read_ref(f'I{i}', ifile)

        # get brightness temperature data
        for i in range(4, 6):
            data[f'bti{i}'] = rh.read_bt(f'I{i}', ifile)

        # get radiance data
        for i in range(1, 6):
            data[f'radi{i}'] = rh.read_rad(f'I{i}', ifile)

    return data


def read_npp_vmaes_l1(path):
    """
        Read VIIRS M-band imagery product (VMAES_L1)
    Args:
        path (string): path to the desired file
    Returns:
        dict: datasets with reflectance, radiance,
            brightness temperature data, geo-reference
            as masked np-arrays
    """
    data = {}
    geo = {}
    with Dataset(path, 'r') as mfile:
        # get reflectance data
        for i in range(1, 12):
            data[f'refm{i}'] = rh.read_ref(f'M{i}', mfile)

        # get brightness temperature data
        for i in range(12, 17):
            data[f'btm{i}'] = rh.read_bt(f'M{i}', mfile)

        # get radiance data
        for i in range(1, 17):
            data[f'radm{i}'] = rh.read_rad(f'M{i}', mfile)

        # get geo data
        geo['lat'] = rh.read_lat(mfile)
        geo['lon'] = rh.read_lon(mfile)
        geo['solza'] = rh.read_solza(mfile)
        geo['solaa'] = rh.read_solaa(mfile)
        geo['satza'] = rh.read_satza(mfile)
        geo['sataa'] = rh.read_sataa(mfile)

    return data, geo


def read_npp_cldmsk_l2(path):
    """
        Read VIIRS Cloud Mask product (CLDMSK_L2)
    Args:
        path (string): path to the desired file
    Returns:
        dict: datasets clear_conf and integer cloud mask as masked np-arrays
    """
    data = {}
    with Dataset(path, 'r') as file:
        data['clear_conf'] = file.groups['geophysical_data'].variables['Clear_Sky_Confidence'][::]
        data['cloud_mask'] = file.groups['geophysical_data'].variables['Integer_Cloud_Mask'][::]
    return data



import subprocess as sp
import os
from datetime import timedelta
from concurrent.futures import ProcessPoolExecutor


def _test_cmrfetch():
    """
        Test for availability of cmrfetch
        Test was unsuccessfull in the case
            of any exception
    """
    cf = 'cmrfetch'
    try:
        sp.run(
            [cf, "--version"], check=True,
            stdout=sp.DEVNULL, stderr=sp.DEVNULL
        )
    except FileNotFoundError:
        raise Exception(
            f'{cf} is not accessible'
        )
    except sp.CalledProcessError as e:
        if e.returncode != 0:
            raise Exception(
                f"{cf} was found, failed to execute"
            )


def _get_data_for_interval(
    path,
    start_d, end_d,
    names,
    geobox,
    dconc=4
):
    """
        Worker function for downloading data itself
    Args:
        path (string): path for saving downloaded files
        start_d (datetime.datetime): start of the
            time selection interval (including)
        end_d (datetime.datetime): end of the
            time selection interval (excluding)
        names (list of string): shortnames for
            desired collections
        geobox (string): bounding box for selecting granules
            for format information check cmrfetch docs
        dconc (int): number of concurrently
            downloading connections
    """
    snames = ''
    for i in names:
        snames += f'-s {i} '
    x = 1
    cnt = 0
    while (x != 0):
        # really robust solution, timeouts are handled by cmrfetch
        # so they cannot be recognized as runtime error below
        # TODO: cmrfetch output analysis for re-downloading failed files
        fetch_cli = f'cmrfetch granules {snames} -t {start_d},{end_d} --bounding-box {geobox} --download {path} --download-concurrency {dconc}'
        r = sp.run(fetch_cli.split())
        x = r.returncode
        cnt += 1
        if (cnt == 5):
            raise RuntimeError(f'Cannot download data for {start_d},{end_d}')


def assimilate(
    names, geobox,
    start_d, end_d,
    path,
    workers, max_queue=0,
    assim_callback=(lambda path: None),
    dconc=4
):
    """
        Performs data assimilation - download it using cmrfetch
            per day intervals and then perform callback function
            on it (for compression purposes mainly)
    Args:
        names (list of string): shortnames for
            desired collections
        geobox (string): bounding box for selecting granules
            for format information check cmrfetch docs
        start_d (datetime.datetime): start of the
            time selection interval (including)
        end_d (datetime.datetime): end of the
            time selection interval (excluding)
        path (string): path for saving downloaded files
        workers (int): max num of the processes
            performing callbacks
        max_queue (int): max num of the queued callback
            tasks, used for downloading throttling
            (for avoiding disk filling)
        assim_callback (function(path)): callback to perform
            on the each set of data per single day
        dconc (int): number of concurrently
            downloading connections
    """
    _test_cmrfetch()

    if (max_queue == 0):
        max_queue = 2 * workers

    if (not os.path.exists(path)):
        os.mkdir(path)

    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = []
        cur_d = end_d
        next_d = end_d + timedelta(days=1)
        while (cur_d >= start_d):
            cur_d_s = cur_d.strftime('%Y-%m-%d')
            next_d_s = next_d.strftime('%Y-%m-%d')
            step_path = os.path.join(path, f'{cur_d_s}')

            _get_data_for_interval(
                step_path,
                cur_d_s, next_d_s,
                names,
                geobox,
                dconc
            )

            task = executor.submit(assim_callback, step_path)
            futures.append(task)

            next_d = cur_d
            cur_d -= timedelta(days=1)

            # Throttling data downloading
            if (len(futures) > max_queue):
                while (len(futures) > max_queue):
                    futures = [f for f in futures if not f.done()]
                    if (len(futures) != 0):
                        futures[0].result()



import numpy as np
import numpy.ma as ma


def _get_masked(data, thr):
    """
        Do masking on the given array for specified threshold
    Args:
        data (list|np.ndarray): given data array
        thr (data values type): threshold value
    Returns:
        (np.ma.MaskedArray): resulting array
    """
    if thr > 0:
        mask = np.where(data >= thr, 1, 0)
    else:
        mask = np.where(data <= thr, 1, 0)
    masked = ma.masked_array(data, mask=mask)
    return masked


def read_so_data(name, file):
    """
        Read from NASA distributed hdf's ans nc's files
            data stored in Scale-Offset model
    Args:
        name (string): name of the desired dataset
        file (string): path to the file
    Returns:
        np.ma.array: dataset from the file
    """
    scale = 1
    offset = 0
    if 'Scale' in file.variables[name].ncattrs():
        scale = file.variables[name].getncattr('Scale')
    if 'Offset' in file.variables[name].ncattrs():
        offset = file.variables[name].getncattr('Offset')
    ref = file.variables[name][:]
    thr = file.variables[name].getncattr('FILL_TEST_VALUE').split('=')[1]
    r = _get_masked(ref, float(thr)) * scale + offset
    return r


# Wrappers for handy extracting different types of data


def read_ref(band, file):
    return read_so_data(f'Reflectance_{band}', file)


def read_rad(band, file):
    return read_so_data(f'Radiance_{band}', file)


def read_bt(band, file):
    return read_so_data(f'BrightnessTemperature_{band}', file)


def read_lat(file):
    return read_so_data('Latitude', file)


def read_lon(file):
    return read_so_data('Longitude', file)


def read_solza(file):
    return read_so_data('SolarZenithAngle', file)


def read_solaa(file):
    return read_so_data('SolarAzimuthAngle', file)


def read_satza(file):
    return read_so_data('SatelliteZenithAngle', file)


def read_sataa(file):
    return read_so_data('SatelliteAzimuthAngle', file)



import xarray as xr
import numpy as np
from math import nan

# Internal functions:


def _check_data(data):
    """
        Check object for being of appropriate type
        Currently supported types:
            * np.ndarray
            * xr.DataArray
    Args:
        data (any): object to be tested
    """
    if not isinstance(data, (np.ndarray, xr.DataArray)):
        raise ValueError(
            "Input data is not array (numpy or xarray) object"
        )


def _get_da(ds, name):
    """
        Helper function for preserving datasets attributes
        that are crucial for later operations
        like saving to the geo-tiff
    Args:
        ds (xr.Dataset): desired dataset
        name (string): xr.DataArray to extract
    Returns:
        xr.DataArray
    """
    da = ds[name]
    attrs_to_copy = ['crs', 'grid_mapping']
    for attr in attrs_to_copy:
        if attr in ds.attrs:
            da.attrs[attr] = ds.attrs[attr]
    return da


# Public functions:


def max_values(data):
    """
        Single function for finding
        max values per time slice
        for various data forms
    Args:
        data (np.ndarray|xr.DataArray):
            slice or set of slices for searching for max values
    Returns
        (xr.DataArray|dtype of data vales):
            max value for given slice or xr.Datarray of such values
            with time dimension
    """
    if isinstance(data, np.ndarray):
        return np.nanmax(data)
    elif isinstance(data, xr.DataArray):
        dims = list(data.dims)
        data_max = None
        if 'time' in dims:
            dims.pop(dims.index('time'))
            data_max = data.max(dim=dims)
        else:
            data_max = data.max(dim=dims).item()
    else:
        raise ValueError(
            "Input data is not array (numpy or xarray) object"
        )
    return data_max


def median_values(data):
    """
        Single function for finding
        median values per time slice
        for various data forms
    Args:
        data (np.ndarray|xr.DataArray):
            slice or set of slices for searching for max values
    Returns
        (xr.DataArray|dtype of data vales):
            median value for given slice or xr.Datarray of such values
            with time dimension
    """
    if isinstance(data, np.ndarray):
        return np.nanmedian(data)
    elif isinstance(data, xr.DataArray):
        dims = list(data.dims)
        data_max = None
        if 'time' in dims:
            dims.pop(dims.index('time'))
            data_max = data.median(dim=dims)
        else:
            data_max = data.median(dim=dims).item()
    else:
        raise ValueError(
            "Input data is not array (numpy or xarray) object"
        )
    return data_max


def nan_mask(data):
    """
        Performs type-independent Nan-checking
    Args:
        data(np.ndarray|xr.DataArray|xr.Dataset):
            object to be tested
    """
    if isinstance(data, np.ndarray):
        return np.isnan(data)
    elif isinstance(data, (xr.DataArray, xr.Dataset)):
        return data.isnull()
    else:
        raise ValueError(
            "Input data is not array (numpy or xarray) object"
        )


def not_nan_mask(data):
    """
        Performs type-independent not-Nan-checking
    Args:
        data(np.ndarray|xr.DataArray|xr.Dataset):
            object to be tested
    """
    if isinstance(data, np.ndarray):
        return np.logical_not(np.isnan(data))
    elif isinstance(data, (xr.DataArray, xr.Dataset)):
        return data.notnull()
    else:
        raise ValueError(
            "Input data is not array (built-in, np or xr) object"
        )


def merge_day_night(day, night, nm):
    """
        Merge day and night composits
    Args:
        day_cm(np.ndarray|xr.DataArray):
            day composit
        night_cm(np.ndarray|xr.DataArray):
            night composit
        nm(np.ndarray|xr.DataArray):
            binary night mask
    Returns:
        (np.ndarray|xr.DataArray):
            merged composit
    """
    mask = xr.where(nm == 0, day, night)
    return mask


def ndvi(NIR, R):
    """
        Get NDVI index from reflectance bands
        Most common use is (I2, I1) or (M7, M5) as (NIR, R)
    Args:
        NIR(np.ndarray|xr.DataArray): near-infrared reflectance band
        R(np.ndarray|xr.DataArray): red reflectance band
    Returns:
        (np.ndarray|xr.DataArray): NDVI index for each pixel,
            can contain NaN values
    """
    i = (NIR - R) / (NIR + R)
    mask = nan_mask(NIR)
    return xr.where(mask, nan, i)



import xarray as xr
from math import nan

from . import Utils


# Internal functions

def _water_bodies_day(ri1, ri2, ri3):
    """
        Day reflectance water bodies test
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
    Returns:
        (np.ndarray|xr.Dataset): binary water bodies mask,
            0 is clear pixel, 1 is water body
            Can contain NaN values
    """
    mask = xr.where(ri1 > ri2, 1, 0)
    mask = xr.where(ri2 > ri3, mask, 0)
    nanmask = Utils.nan_mask(ri1)
    mask = xr.where(nanmask, nan, mask)
    return mask


# Public wrappers:


def water_bodies_day(ri1, ri2, ri3):
    """
        Day reflectance water bodies test
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
    Returns:
        (np.ndarray|xr.Dataset): binary water bodies mask,
            0 is clear pixel, 1 is water body
            Can contain NaN values
    """
    Utils._check_data(ri1)
    Utils._check_data(ri2)
    Utils._check_data(ri3)
    return _water_bodies_day(ri1, ri2, ri3)


# Public xr.Dataset wrappers:


def water_bodies_day_ds(ds):
    """
        wrapper for day reflectance water bodies test
            for xr.Dataset objects
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
    Returns:
        (np.ndarray|xr.Dataset): binary water bodies mask,
            0 is clear pixel, 1 is water body
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return water_bodies_day(
        ds['I01'],
        ds['I02'],
        ds['I03']
    )



import xarray as xr
from math import nan

from . import Utils


# Internal functions

def _active_fires(ri1, ri2, ri3, bi4, bi5, nmask, cmask, wmask=None):
    """
        Fire mask from actives fire VIIRS product
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
        nmask (np.ndarray|xr.Dataset): Day/night mask (1 is night)
        cmask (np.ndarray|xr.Dataset): Cloud mask (1 is clear pixel)
        wmask (np.ndarray|xr.Dataset, optional): water bodies mask,
            (1 is water body)
    Returns:
        (np.ndarray|xr.Dataset): binary fire mask,
            0 is clear pixel, 1 is fire
            Can contain NaN values
    """
    ri12 = ri1 + ri2
    bi45 = bi4 - bi5
    saturated_bi4 = xr.where(bi4 == 367, 1, 0)

    # Fixed threshold tests:
    # night tests
    nt0 = xr.where(bi4 > 320, 1, 0)

    nt1 = xr.where(bi45 < 0, 1, 0)
    nt1_1 = xr.where(bi5 > 310, 1, 0)
    nt1_2 = xr.where(bi4 == 208, 1, 0)
    nt1_2 = xr.where(bi5 > 335, nt1_2, 0)
    nt1 = xr.where((nt1_1 + nt1_2) != 0, nt1, 0)

    # day tests
    dt0 = xr.where(saturated_bi4, 1, 0)
    dt0 = xr.where(bi5 > 290, dt0, 0)
    dt0 = xr.where(ri12 > 70, dt0, 0)

    dt1 = xr.where(bi45 < 0, 1, 0)
    dt1 = xr.where(bi5 > 325, dt1, 0)

    # Potential background fires (to be masked):
    # night tests
    nt2 = xr.where(bi4 > 335, 0, 1)
    nt2 = xr.where(bi4 > 30, 0, nt2)

    # day tests
    dt2 = xr.where(bi4 > 300, 0, 1)
    dt2 = xr.where(bi45 > 10, 0, dt2)

    # Bright fire-free targets (daytime only):
    ft = xr.where(ri12 > 60, 1, 0)
    ft = xr.where(bi5 < 285, ft, 0)
    ft = xr.where(ri3 > 30, ft, 0)
    ft = xr.where(ri3 > ri2, ft, 0)
    ft = xr.where(ri2 > 25, ft, 0)
    ft = xr.where(bi4 <= 235, ft, 0)

    # # Candidate fire pixels:
    # # night test
    # cnt = xr.where(bi4 > 295, 1, 0)
    # cnt = xr.where(bi45 > 10, 1, cnt)
    #
    # # day test
    # cdt = xr.where(bi45 > 25, 1, 0)
    # M = Utils.median_values(bi4)
    # bi4m = xr.where(M > 325, M, 325)
    # bi4s = xr.where(bi4m < 330, bi4m, 330)
    # cdt = xr.where(bi4 > bi4s, 1, cdt)

    # Results merging:
    # night
    nt = (nt0 + nt1) * nt2  # + cnt
    nt = xr.where(nt > 0, 1, 0)

    # day
    dt = (dt0 + dt1) * dt2  # + cdt
    dt = xr.where(dt > 0, 1, 0)
    dt = xr.where(ft == 1, 0, dt)

    # merge
    mask = xr.where(nmask == 1, nt, nan)
    mask = xr.where(nmask == 0, dt, mask)
    mask = xr.where(cmask == 0, nan, mask)
    if wmask is not None:
        mask = xr.where(wmask == 0, nan, mask)
    return mask


# Public wrappers:


def active_fires(ri1, ri2, ri3, bi4, bi5, nmask, cmask, wmask=None):
    """
        Fire mask from actives fire VIIRS product
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
        nmask (np.ndarray|xr.Dataset): Day/night mask (1 is night)
        cmask (np.ndarray|xr.Dataset): Cloud mask (1 is clear pixel)
        wmask (np.ndarray|xr.Dataset, optional): water bodies mask,
            (1 is water body)
    Returns:
        (np.ndarray|xr.Dataset): binary fire mask,
            0 is clear pixel, 1 is fire
            Can contain NaN values
    """
    Utils._check_data(ri1)
    Utils._check_data(ri2)
    Utils._check_data(ri3)
    Utils._check_data(bi4)
    Utils._check_data(bi5)
    Utils._check_data(nmask)
    Utils._check_data(cmask)
    if wmask is not None:
        Utils._check_data(wmask)
    return _active_fires(
        ri1, ri2, ri3,
        bi4, bi5,
        nmask, cmask, wmask
    )


# Public xr.Dataset wrappers:


def active_fires_ds(ds, nmask, cmask, wmask=None):
    """
        Wrapper for fire mask from actives fire VIIRS product
            for xr.Dataset objects
        Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
        The New VIIRS 375 m active fire detection data product:
            Algorithm description and initial assessment
    Args:
        ri1 (np.ndarray|xr.Dataset): I01 in reflectance calibration
        ri2 (np.ndarray|xr.Dataset): I02 in reflectance calibration
        ri3 (np.ndarray|xr.Dataset): I03 in reflectance calibration
        bi4 (np.ndarray|xr.Dataset): I04 in BT calibration
        bi5 (np.ndarray|xr.Dataset): I05 in BT calibration
        nmask (np.ndarray|xr.Dataset): Day/night mask (1 is night)
        cmask (np.ndarray|xr.Dataset): Cloud mask (1 is clear pixel)
        wmask (np.ndarray|xr.Dataset, optional): water bodies mask,
            (1 is water body)
    Returns:
        (np.ndarray|xr.Dataset): binary fire mask,
            0 is clear pixel, 1 is fire
            Can contain NaN values
    """
    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            "Incorrect input data format"
        )
    return active_fires(
        ds['I01'],
        ds['I02'],
        ds['I03'],
        ds['I04'],
        ds['I05'],
        nmask, cmask,
        wmask=wmask
    )



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

with open('LICENSE', 'r') as f:
    license_text = f.read()

setup(
    name='viirs-tools',
    description='Python library for processing VIIRS data',
    long_description_content_type='text/markdown',
    long_description=long_description,
    url='https://github.com/Veon2479/viirs-tools',
    license=license_text,
    author='Andrey Shuliak',
    version='1.0.2',
    packages=find_packages(),
    install_requires=[
        'xarray',
    ],
    extras_require={
        'assimilator': ['netcdf4']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Atmospheric Science'
    ],
    keywords=['viirs', 'satellite', 'remote sensing', 'meteorology'],

)
